#!/usr/bin/env python
# Rui Zhang 8.2020
# rui.zhang@cern.ch

import math
import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
from torch.autograd import Function

class BinarizeF(Function):
    @staticmethod
    def forward(cxt, input):
        output = input.new(input.size())
        output[input >= 0] = 1
        output[input < 0] = -1
        return output
    @staticmethod
    def backward(cxt, grad_output):
        grad_input = grad_output.clone()
        return grad_input

class ClipF(Function):
    @staticmethod
    def forward(ctx, input):
        output = input.clone().detach()
        output[input >= 1] = 1
        output[input <= 0] = 0
        ctx.save_for_backward(input)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        grad_input = grad_output.clone()
        grad_input[input >= 1] = 0
        grad_input[input <= 0] = 0
        return grad_input
class QF_FB_NC(nn.Linear):
    def sim_neural_comp(self, input_ori, w_ori):
        p = input_ori
        d = 4 * p * (1 - p)
        e = (2 * p - 1)
        w = w_ori
        sum_of_sq = (d + e.pow(2)).sum(-1)
        #sum_of_sq = (e.pow(2)).sum(-1)
        sum_of_sq = sum_of_sq.unsqueeze(-1)
        sum_of_sq = sum_of_sq.expand(p.shape[0], w.shape[0])
        diag_p = torch.diag_embed(e)
        p_w = torch.matmul(w, diag_p)
        z_p_w = torch.zeros_like(p_w)
        shft_p_w = torch.cat((p_w, z_p_w), -1)
        sum_of_cross = torch.zeros_like(p_w)
        length = p.shape[1]
        for shft in range(1, length):
            sum_of_cross += shft_p_w[:, :, 0:length] * shft_p_w[:, :, shft:length + shft]
        sum_of_cross = sum_of_cross.sum(-1)
        result = (sum_of_sq + 2 * sum_of_cross) / (length ** 2) # Equation (4)
        return result

    def forward(self, input):
        binarize = BinarizeF.apply
        binary_weight = binarize(self.weight)
        if self.bias is None:
            return self.sim_neural_comp(input, binary_weight)
        else:
            print("Bias is not supported at current version")
            sys.exit(0)

    def reset_parameters(self):
        in_features, out_features = self.weight.size()
        stdv = math.sqrt(1.5 / (in_features + out_features))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.zero_()
        self.weight.lr_scale = 1. / stdv

class QF_FB_BN_IAdj(nn.Module):
    def __init__(self, num_features, init_ang_inc=1, momentum=0.1,training = False):
        super(QF_FB_BN_IAdj, self).__init__()

        self.x_running_rot = Parameter(torch.zeros((num_features)), requires_grad=False)
        self.ang_inc = Parameter(torch.tensor(init_ang_inc,dtype=torch.float32),requires_grad=True)
        self.momentum = momentum

        self.printed = False
        self.x_mean_ancle = 0
        self.x_mean_rote = 0
        self.input = 0
        self.output = 0

    def forward(self, x, training=True):
        if not training:
            if not self.printed:
                self.printed = True
            x_1 = (self.x_running_rot * x)

        else:
            self.printed = False
            x = x.transpose(0, 1)
            x_sum = x.sum(-1).unsqueeze(-1).expand(x.shape)
            x_lack_sum = x_sum + x
            x_mean = x_lack_sum / x.shape[-1]
            ang_inc = (self.ang_inc > 0).float() * self.ang_inc + 1
            y = 0.5 / x_mean
            y = y.transpose(0, 1)
            y = y / ang_inc
            y = y.transpose(0, 1)
            x_moving_rot = (y.sum(-1) / x.shape[-1])
            self.x_running_rot[:] = self.momentum * self.x_running_rot + \
                                    (1 - self.momentum) * x_moving_rot
            x_1 = y * x
            x_1 = x_1.transpose(0, 1)
        return x_1

    def reset_parameters(self):
        self.reset_running_stats()
        self.ang_inc.data.zeros_()


class QF_FB_BN_BAdj(nn.Module):
    def __init__(self, num_features, momentum=0.1):
        super(QF_FB_BN_BAdj, self).__init__()
        self.x_running_rot = Parameter(torch.zeros(num_features), requires_grad=False)
        self.momentum = momentum
        self.x_l_0_5 = Parameter(torch.zeros(num_features), requires_grad=False)
        self.x_g_0_5 = Parameter(torch.zeros(num_features), requires_grad=False)

    def forward(self, x, training=True):
        if not training:
            x_1 = self.x_l_0_5 * (self.x_running_rot * (1 - x) + x)
            x_1 += self.x_g_0_5 * (self.x_running_rot * x)
        else:
            x = x.transpose(0, 1)
            x_sum = x.sum(-1)
            x_mean = x_sum / x.shape[-1]
            self.x_l_0_5[:] = ((x_mean <= 0.5).float())
            self.x_g_0_5[:] = ((x_mean > 0.5).float())
            y = self.x_l_0_5 * ((0.5 - x_mean) / (1 - x_mean))
            y += self.x_g_0_5 * (0.5 / x_mean)

            self.x_running_rot[:] = self.momentum * self.x_running_rot + \
                                    (1 - self.momentum) * y
            x = x.transpose(0, 1)
            x_1 = self.x_l_0_5 * (y * (1 - x) + x)
            x_1 += self.x_g_0_5 * (y * x)
        return x_1

class Net(nn.Module):
    def __init__(self,in_size,layers=[3],with_norm=True,given_ang=[[-1, 1, 1]],training=False,binary=False,classic=False):
        super(Net, self).__init__()

        self.in_size = in_size
        self.training = training
        self.with_norm = with_norm
        self.layer = len(layers)
        self.binary = binary
        self.classic = classic
        cur_input_size = in_size
        for idx in range(self.layer):
            fc_name = "fc"+str(idx)
            if classic:
                setattr(self, fc_name, MLP(cur_input_size, layers[idx], bias=False))
            else:
                setattr(self, fc_name, QF_FB_NC(cur_input_size, layers[idx], bias=False))
            cur_input_size = layers[idx]

        if self.with_norm:
            for idx in range(self.layer):
                IAdj_name = "IAdj"+str(idx)
                BAdj_name = "BAdj"+str(idx)
                setattr(self, BAdj_name, QF_FB_BN_BAdj(num_features=layers[idx]))
                setattr(self, IAdj_name, QF_FB_BN_IAdj(num_features=layers[idx], init_ang_inc=given_ang[idx], training=training))
            for idx in range(self.layer):
                bn_name = "bn"+str(idx)
                setattr(self, bn_name,nn.BatchNorm1d(num_features=layers[idx]))


    def forward(self, x, training=1):
        binarize = BinarizeF.apply
        clipfunc = ClipF.apply
        x = x.view(-1, self.in_size)
        if self.classic == 1 and self.with_norm==0:
            for layer_idx in range(self.layer):
                if self.binary and layer_idx==0:
                    x = binarize(x-0.5)
                x = getattr(self, "fc" + str(layer_idx))(x)
                x = x.pow(2)
        elif self.classic == 1 and self.with_norm==1:
            for layer_idx in range(self.layer):
                if self.binary and layer_idx==0:
                    x = (binarize(x - 0.5) + 1) / 2
                x = getattr(self, "fc" + str(layer_idx))(x)
                x = x.pow(2)
                x = getattr(self, "bn" + str(layer_idx))(x)
                x = clipfunc(x)

        elif self.classic == 0 and self.with_norm==0:
            for layer_idx in range(self.layer):
                if self.binary and layer_idx==0:
                    x = (binarize(x - 0.5) + 1) / 2
                x = getattr(self, "fc" + str(layer_idx))(x)
        else:   # Quantum Training
            if self.training == 1:
                for layer_idx in range(self.layer):
                    if self.binary and layer_idx==0:
                        x = (binarize(x-0.5)+1)/2
                    x = getattr(self, "fc"+str(layer_idx))(x)
                    x = getattr(self, "BAdj"+str(layer_idx))(x)
                    x = getattr(self, "IAdj"+str(layer_idx))(x)
            else:
                for layer_idx in range(self.layer):
                    if self.binary and layer_idx==0:
                        x = (binarize(x-0.5)+1)/2
                    x = getattr(self, "fc"+str(layer_idx))(x)
                    x = getattr(self, "BAdj"+str(layer_idx))(x, training=False)
                    x = getattr(self, "IAdj"+str(layer_idx))(x, training=False)
        return x




#!/usr/bin/env python
# Rui Zhang 8.2020
# rui.zhang@cern.ch

# import numpy as np
# import torch
# from torch.autograd import Function

# class HybridFunction(Function):
#     """ Hybrid quantum - classical function definition """
    
#     @staticmethod
#     def forward(ctx, input, quantum_circuit, shift):
#         """ Forward pass computation """
#         ctx.shift = shift
#         ctx.quantum_circuit = quantum_circuit

#         print('zhangr input forward', input, input.tolist())
#         expectation_z = ctx.quantum_circuit.run(input.tolist())
#         print('zhangr expectation_z', expectation_z)
#         result = torch.tensor([expectation_z])
#         ctx.save_for_backward(input, result)

#         return result
        
#     @staticmethod
#     def backward(ctx, grad_output):
#         """ Backward pass computation """
#         input, expectation_z = ctx.saved_tensors
#         input_list = np.array(input.tolist())
        
#         shift_right = input_list + np.ones(input_list.shape) * ctx.shift
#         shift_left = input_list - np.ones(input_list.shape) * ctx.shift
        
#         gradients = []
#         for i in range(len(input_list)):
#             expectation_right = ctx.quantum_circuit.run(shift_right[i]) / (2 * ctx.shift)
#             expectation_left  = ctx.quantum_circuit.run(shift_left[i]) / (2 * ctx.shift)
            
#             gradient = torch.tensor([expectation_right]) - torch.tensor([expectation_left])
#             gradients.append(gradient)
#         gradients = np.array([gradients]).T
#         return torch.tensor([gradients]).float() * grad_output.float(), None, None



#!/usr/bin/env python
# Rui Zhang 8.2020
# rui.zhang@cern.ch

import logging
import numpy as np
import matplotlib.pyplot as plt
import importlib

import torch
from torch.utils.data import Dataset, DataLoader
from torchvision import datasets, transforms
import torch.optim as optim
import torch.nn as nn
# from torchviz import make_dot
from torchsummary import summary
import qiskit
from qiskit.visualization import circuit_drawer

from quhep.train.train import _train, add_options


# https://discuss.pytorch.org/t/input-numpy-ndarray-instead-of-images-in-a-cnn/18797/2
class MyDataset(Dataset):
    ''' Transform numpy array to pyTorch Dataset '''
    def __init__(self, data, target, transform=None):
        self.data = torch.from_numpy(data).float()
        self.target = torch.from_numpy(target).long()
        self.transform = transform
        
    def __getitem__(self, index):
        x = self.data[index]
        y = self.target[index]
        
        if self.transform:
            x = self.transform(x)
        
        return x, y

    def __len__(self):
        return len(self.data)


class Train(_train):
    def describe(self): return self.__class__.__name__
    def __init__(self, **kargs):
        self.m_use_quple = kargs['use_quple']
        self.m_nlayers = kargs['nlayers']
        self.m_ndepths = kargs['ndepths']
        self.m_encoder_type = kargs['encoder_type']
        self.m_Qnetwork = kargs['Qnetwork']
        self.m_loss = kargs['loss']
        quple_par = f'use_qsk-enc{self.m_encoder_type}-Dep{self.m_ndepths}-QNN{self.m_Qnetwork}-L{self.m_nlayers}-l{self.m_loss}'
        super().__init__(additiona_log=quple_par, **kargs)

        self.m_scale_range = (0, 1)

    def train_quantum_model(self):
        dataset_train = MyDataset(self.m_data.x_train, self.m_data.y_train)
        dataset_test = MyDataset(self.m_data.x_test, self.m_data.y_test)
        dataset_val = MyDataset(self.m_data.x_val, self.m_data.y_val)
        loader_train = DataLoader(dataset_train, batch_size = self.m_batch_size, shuffle = False, num_workers = 0, pin_memory = torch.cuda.is_available())
        loader_test = DataLoader(dataset_test, batch_size = self.m_batch_size, shuffle = False, num_workers = 0, pin_memory = torch.cuda.is_available())

        module = importlib.import_module(f'quhep.qsk_circ.QuantumFlow')
        model = module.Net(self.m_nqubits)
        # model.hybrid.plot_circuit(f'{self.m_log_dir["model"]}/circuit') #TODO
        # summary(model, input_size=(self.m_nqubits, 1))
        optimizer = optim.Adam(model.parameters(), lr=0.001)
        loss_func = nn.NLLLoss()

        loss_list = []

        model.train()
        for epoch in range(self.m_epochs):
            total_loss = []
            for batch_idx, (data, target) in enumerate(loader_train):
                optimizer.zero_grad()
                # Forward pass
                output = model(data)
                # Calculating loss
                loss = loss_func(output, target)
                # Backward pass
                loss.backward()
                # Optimize the weights
                optimizer.step()
                
                total_loss.append(loss.item())
            loss_list.append(sum(total_loss)/len(total_loss))
            print('Training [{:.0f}%]\tLoss: {:.4f}'.format(100. * (epoch + 1) / self.m_epochs, loss_list[-1]))

        plt.plot(loss_list)
        plt.title('Hybrid NN Training Convergence')
        plt.xlabel('Training Iterations')
        plt.ylabel('Neg Log Likelihood Loss')


        model.eval()
        with torch.no_grad():
            correct = 0
            for batch_idx, (data, target) in enumerate(loader_test):
                output = model(data)
            
                pred = output.argmax(dim=1, keepdim=True) 
                correct += pred.eq(target.view_as(pred)).sum().item()
            
                loss = loss_func(output, target)
                total_loss.append(loss.item())
            
            print('Performance on test data:\n\tLoss: {:.4f}\n\tAccuracy: {:.1f}%'.format(sum(total_loss) / len(total_loss), correct / len(loader_test) * 100))


def main():
    args = add_options()
    if args.fix_seed:
        import numpy as np
        np.random.seed(args.fix_seed + 666)

    trainer = Train(**vars(args))
    trainer.get_classical_data()

    trainer.train_quantum_model()
    if args.run_classical:
        trainer.train_classical_model()
        trainer.plotROC()
    trainer.plotROC(True)
    logging.info(f'Job finished')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Rui Zhang 8.2020
# rui.zhang@cern.ch

import numpy as np
import qiskit
import torch
from torch.autograd import Function

class QuantumCircuit:
    """ 
    This class provides a simple interface for interaction 
    with the quantum circuit 
    """
    
    def __init__(self, n_qubits, backend, shots):
        # --- Circuit definition ---
        print('zhangr QuantumCircuit init')
        self._circuit = qiskit.QuantumCircuit(n_qubits)
        
        all_qubits = [i for i in range(n_qubits)]
        self.parameters = [qiskit.circuit.Parameter(f'θ_{i}') for i in range(n_qubits)]
        # self.parameters = [qiskit.circuit.Parameter(f'θ_0'), qiskit.circuit.Parameter(f'θ_1')]
        print('zhangr self.parameters', self.parameters[0], n_qubits)
        
        self._circuit.h(all_qubits)
        self._circuit.barrier()
        for theta, qubit in zip(self.parameters, all_qubits):
            self._circuit.ry(theta, qubit)
        # self._circuit.ry(self.parameters[0], all_qubits[0])
        # self._circuit.ry(self.parameters[1], all_qubits[1])
        
        self._circuit.measure_all()
        # ---------------------------

        self.backend = backend
        self.shots = shots
    
    def run(self, thetas):
        print('zhangr self.parameters', self.parameters[0])
        print('zhangr thetas', thetas)
        print('zhangr dict(zip(self.parameters, thetas)))', [{self.parameters[i]: thetas[0][i]} for i in range(2)])

        total_results = {}
        for theta_event in thetas:
            print('zhangr theta_event', theta_event)
            job = qiskit.execute(self._circuit, 
                                self.backend, 
                                shots = self.shots,
                                parameter_binds = [dict(zip(self.parameters, theta_event)) ])
            result = job.result().get_counts(self._circuit)
            total_results = {x: total_results.get(x, 0) + result.get(x, 0) for x in set(total_results).union(result)}
            print('zhangr result', result)
            print('zhangr total_results', total_results)
        
        counts = np.array(list(result.values()))
        states = np.array(list(result.keys())).astype(float)
        
        # Compute probabilities for each state
        probabilities = counts / self.shots
        # Get state expectation
        print('zhangr states', states)
        print('zhangr probabilities', probabilities)
        expectation = np.sum(states * probabilities)
        
        return np.array([expectation])

class HybridFunction(Function):
    """ Hybrid quantum - classical function definition """
    
    @staticmethod
    def forward(ctx, input, quantum_circuit, shift):
        """ Forward pass computation """
        ctx.shift = shift
        ctx.quantum_circuit = quantum_circuit

        print('zhangr input forward', input, input.tolist())
        expectation_z = ctx.quantum_circuit.run(input.tolist())
        print('zhangr expectation_z', expectation_z)
        result = torch.tensor([expectation_z])
        ctx.save_for_backward(input, result)

        return result
        
    @staticmethod
    def backward(ctx, grad_output):
        """ Backward pass computation """
        input, expectation_z = ctx.saved_tensors
        input_list = np.array(input.tolist())
        
        shift_right = input_list + np.ones(input_list.shape) * ctx.shift
        shift_left = input_list - np.ones(input_list.shape) * ctx.shift
        
        gradients = []
        for i in range(len(input_list)):
            expectation_right = ctx.quantum_circuit.run(shift_right[i]) / (2 * ctx.shift)
            expectation_left  = ctx.quantum_circuit.run(shift_left[i]) / (2 * ctx.shift)
            
            gradient = torch.tensor([expectation_right]) - torch.tensor([expectation_left])
            gradients.append(gradient)
        gradients = np.array([gradients]).T
        return torch.tensor([gradients]).float() * grad_output.float(), None, None


class Hybrid(nn.Module):
    """ Hybrid quantum - classical layer definition """
    
    def __init__(self, nqubits, backend, shots, shift):
        super(Hybrid, self).__init__()
        # module = importlib.import_module(f'quhep.qsk_circ.QuantumCircuit')
        self.quantum_circuit = QuantumCircuit(nqubits, backend, shots)
        self.shift = shift
        
    def forward(self, input):
        # module = importlib.import_module(f'quhep.qsk_prop.HybridFunction')
        return HybridFunction.apply(input, self.quantum_circuit, self.shift)

    def circuit(self):
        return self.quantum_circuit

    def plot_circuit(self, filename):
        if filename.endswith('.pdf') or filename.endswith('.txt'):
            filename = filename[:-4]
        circuit_drawer(self.quantum_circuit._circuit, filename=f'{filename}.pdf', output='mpl')
        logging.info(f'\n{self.quantum_circuit._circuit.draw()}')
        logging.info(f'Circuits stored in: f"{filename}.pdf"')

class Net(nn.Module):
    def __init__(self, nqubits, shots = 1000, shift = 0.05):
        super(Net, self).__init__()
        self.m_nqubits = nqubits
        self.input = nn.Linear(self.m_nqubits, self.m_nqubits)
        self.hybrid = Hybrid(self.m_nqubits, qiskit.Aer.get_backend('qasm_simulator'), shots, shift)
        self.output = nn.Linear(self.m_nqubits, 1)

    def forward(self, x):
        print('zhangr ax input', x.shape, x)
        x = self.input(x)
        print('zhangr bx after input', x.shape, x)
        x = self.QF_FB_NC(x)
        print('zhangr cx after QF_FB_NC', x.shape, x)
        x = self.hybrid(x)
        print('zhangr cx after hybrid', x.shape, x)
        # x = self.output(x.float())
        # print('zhangr dx after output', x.shape, x)
        x = torch.sigmoid(x)
        return x
